"""
To evaluate the performance of the solar forecast, a predefined testset is used.

A file has been added to this branch (make-testset) which defines a set of
random timestamps and sites ids.
This contains 50 sites each with 50 timestamps to make 2500 samples in total.

"""

import os

import pandas as pd
from dotenv import load_dotenv
from huggingface_hub.hf_api import HfFolder

from quartz_solar_forecast.eval.forecast import run_forecast
from quartz_solar_forecast.eval.metrics import metrics
from quartz_solar_forecast.eval.nwp import get_nwp
from quartz_solar_forecast.eval.pv import get_pv_metadata, get_pv_truth
from quartz_solar_forecast.eval.utils import combine_forecast_ground_truth

load_dotenv()

try:
    hf_token = os.environ["HF_TOKEN"]
    HfFolder.save_token(hf_token)
except Exception:
    print(
        "Warning, you wont be able to run evaluation if you dont set your "
        "Hugging Face Access Token to HF_TOKEN, or be logged in with Hugging Face"
    )


def run_eval(testset_path: str = "dataset/testset.csv"):
    # load testset from csv
    testset = pd.read_csv(testset_path)

    # Extract generation data and metadata for specific sites and timestamps
    # for the testset from Hugging Face. (Zak)
    pv_metadata = get_pv_metadata(testset)

    # Split data into PV inputs and ground truth. (Zak)
    ground_truth_df = get_pv_truth(testset)

    # Collect NWP data from Hugging Face, ICON. (Peter)
    nwp_df = get_nwp(pv_metadata)

    # Run forecast with PV and NWP inputs.
    predictions_df = run_forecast(pv_df=pv_metadata, nwp_df=nwp_df)

    # Combine the forecast results with the ground truth
    # (ts, id, horizon (in hours), pred, truth, diff)
    results_df = combine_forecast_ground_truth(predictions_df, ground_truth_df)

    # Save file
    results_df.to_csv("results.csv")

    # Calculate and print metrics: MAE
    metrics(results_df, pv_metadata, include_night=True)
    metrics(results_df, pv_metadata, include_night=False)

    # Visualizations
    # TODO


# run_eval()



from datetime import datetime

from pydantic import BaseModel, Field

from quartz_solar_forecast.inverters.enphase import EnphaseInverter, EnphaseSettings
from quartz_solar_forecast.inverters.givenergy import GivEnergyInverter, GivEnergySettings
from quartz_solar_forecast.inverters.mock import MockInverter
from quartz_solar_forecast.inverters.solarman import SolarmanInverter, SolarmanSettings
from quartz_solar_forecast.inverters.solis import SolisInverter, SolisSettings
from quartz_solar_forecast.inverters.victron import VictronInverter, VictronSettings


class PVSite(BaseModel):
    latitude: float = Field(
        ...,
        description="the latitude of the site",
        ge=-90,
        le=90,
    )
    longitude: float = Field(
        ...,
        description="the longitude of the site",
        ge=-180,
        le=180,
    )
    capacity_kwp: float = Field(
        ...,
        description="the capacity [kwp] of the site",
        ge=0,
    )
    tilt: float = Field(
        default=35,
        description=(
            "the tilt of the site [degrees], the panels' angle relative to horizontal ground"
        ),
        ge=0,
        le=90,
    )
    orientation: float = Field(
        default=180,
        description=(
            "the orientation of the site [degrees], the angle between north and "
            "the direction the panels face, measured on the horizontal plane."
        ),
        ge=0,
        le=360,
    )

    def round_latitude_and_longitude(self):
        """Round the latitude and longitude to 2 decimal places

        This is to ensure that the location of the site is not stored exactly.
        """
        self.latitude = round(self.latitude, 2)
        self.longitude = round(self.longitude, 2)


class PVSiteWithInverter(PVSite):
    inverter_type: str = Field(
        default=None,
        description="The type of inverter used",
        json_schema_extra=["enphase", "solis", "givenergy", "solarman", None],
    )

    def get_inverter(self):
        if self.inverter_type == "enphase":
            return EnphaseInverter(EnphaseSettings())
        elif self.inverter_type == "solis":
            return SolisInverter(SolisSettings())
        elif self.inverter_type == "givenergy":
            return GivEnergyInverter(GivEnergySettings())
        elif self.inverter_type == "solarman":
            return SolarmanInverter(SolarmanSettings())
        elif self.inverter_type == "victron":
            return VictronInverter.from_settings(VictronSettings())
        else:
            return MockInverter()


class ForecastRequest(BaseModel):
    site: PVSite
    timestamp: datetime | None = None


class TokenRequest(BaseModel):
    redirect_url: str



"""Log usage of this package to Sentry"""

import importlib.metadata
import os

import sentry_sdk

from quartz_solar_forecast.pydantic_models import PVSite

version = importlib.metadata.version("quartz_solar_forecast")

quartz_solar_forecast_logging = (
    os.getenv("QUARTZ_SOLAR_FORECAST_LOGGING", "True").lower() != "false"
)

SENTRY_DSN = "https://b2b6f3c97299f81464bc16ad0d516d0b@o400768.ingest.us.sentry.io/4508439933157376"
sentry_sdk.init(dsn=SENTRY_DSN, traces_sample_rate=1.0)


def write_sentry(params):
    """
    Log usage of this package to Sentry
    """

    if not quartz_solar_forecast_logging:
        return

    try:
        for key, value in params.items():
            # we want to make sure we don't store the exact location of the site
            if isinstance(value, PVSite):
                value.round_latitude_and_longitude()

            # set sentry tag
            sentry_sdk.set_tag(key, value)

        if os.getenv("PYTEST_CURRENT_TEST") is not None:
            sentry_sdk.set_tag("CI Test", "True")

        sentry_sdk.set_tag("version", version)

    except Exception as _:  # noqa
        pass



"""Function to get NWP data and create fake PV dataset"""

import ssl
from datetime import datetime

import numpy as np
import openmeteo_requests
import pandas as pd
import requests_cache
import xarray as xr
from retry_requests import retry

from quartz_solar_forecast.pydantic_models import PVSite, PVSiteWithInverter

ssl._create_default_https_context = ssl._create_unverified_context


def get_nwp(site: PVSite, ts: datetime, nwp_source: str = "icon") -> xr.Dataset:
    """
    Get GFS NWP data for a point time space and time

    :param site: the PV site
    :param ts: the timestamp for when you want the forecast for
    :param nwp_source: the nwp data source. Either "gfs", "icon" or "ukmo". Defaults to "icon"
    :return: nwp forecast in xarray
    """
    now = datetime.now()

    # Setup the Open-Meteo API client with cache and retry on error
    cache_session = requests_cache.CachedSession(".cache", expire_after=-1)
    retry_session = retry(cache_session, retries=5, backoff_factor=0.2)
    openmeteo = openmeteo_requests.Client(session=retry_session)

    # Define the variables we want. Visibility is handled separately after the main request
    variables = [
        "temperature_2m",
        "precipitation",
        "cloud_cover_low",
        "cloud_cover_mid",
        "cloud_cover_high",
        "wind_speed_10m",
        "shortwave_radiation",
        "direct_radiation",
    ]

    start = ts.date()
    end = start + pd.Timedelta(days=7)

    url = ""

    # check whether the time stamp is more than 3 months in the past
    if (now - ts).days > 90:
        print(
            "Warning: The requested timestamp is more than 3 months in the past. "
            "The weather data are provided by a reanalyse model and not ICON or GFS."
        )

        # load data from open-meteo Historical Weather API
        url = "https://archive-api.open-meteo.com/v1/archive"

    else:
        # Getting NWP from open meteo weather forecast API by ICON, GFS, or UKMO
        # within the last 3 months
        url_nwp_source = {"icon": "dwd-icon", "gfs": "gfs", "ukmo": "ukmo_seamless"}.get(nwp_source)
        if not url_nwp_source:
            raise Exception(f'Source ({nwp_source}) must be either "icon", "gfs", or "ukmo"')
        url = (
            f"https://api.open-meteo.com/v1/"
            f"{url_nwp_source if nwp_source != 'ukmo' else 'forecast'}"
        )

    params = {
        "latitude": site.latitude,
        "longitude": site.longitude,
        "start_date": f"{start}",
        "end_date": f"{end}",
        "hourly": variables,
    }

    # Add the "models" parameter if using "ukmo"
    if nwp_source == "ukmo":
        params["models"] = "ukmo_seamless"

    # Make API call to URL
    response = openmeteo.weather_api(url, params=params)
    hourly = response[0].Hourly()

    hourly_data = {
        "time": pd.date_range(
            start=pd.to_datetime(hourly.Time(), unit="s", utc=False),
            end=pd.to_datetime(hourly.TimeEnd(), unit="s", utc=False),
            freq=pd.Timedelta(seconds=hourly.Interval()),
            inclusive="left",
        )
    }

    # variables index as in the variables array of the request
    for idx, var in enumerate(["t", "prate", "lcc", "mcc", "hcc", "si10", "dswrf", "dlwrf"]):
        hourly_data[var] = hourly.Variables(idx).ValuesAsNumpy()

    # handle visibility
    if (now - ts).days <= 90:
        # load data from open-meteo gfs model
        params = {
            "latitude": site.latitude,
            "longitude": site.longitude,
            "start_date": f"{start}",
            "end_date": f"{end}",
            "hourly": "visibility",
        }
        data_vis_gfs = (
            openmeteo.weather_api("https://api.open-meteo.com/v1/gfs", params=params)[0]
            .Hourly()
            .Variables(0)
            .ValuesAsNumpy()
        )
        hourly_data["vis"] = data_vis_gfs
    else:
        # set to maximum visibility possible
        hourly_data["vis"] = 24000.0

    df = pd.DataFrame(data=hourly_data).set_index("time").astype("float64")

    # convert data into xarray
    data_xr = format_nwp_data(df, nwp_source, site)

    return data_xr


def format_nwp_data(df: pd.DataFrame, nwp_source: str, site: PVSite):
    data_xr = xr.DataArray(
        data=df.values,
        dims=["step", "variable"],
        coords={
            "step": ("step", df.index - df.index[0]),
            "variable": df.columns,
        },
    )
    data_xr = data_xr.to_dataset(name=nwp_source)
    data_xr = data_xr.assign_coords(
        {"x": [site.longitude], "y": [site.latitude], "time": [df.index[0]]}
    )
    return data_xr


def process_pv_data(
    live_generation_kw: pd.DataFrame | None, ts: pd.Timestamp, site: "PVSite"
) -> xr.Dataset:
    """
    Process PV data and create an xarray Dataset.

    :param live_generation_kw: DataFrame containing live generation data, or None
    :param ts: Current timestamp
    :param site: PV site information
    :return: xarray Dataset containing processed PV data
    """
    if live_generation_kw is not None and not live_generation_kw.empty:
        # Get the most recent data
        recent_pv_data = live_generation_kw[live_generation_kw["timestamp"] <= ts]
        power_kw = np.array([recent_pv_data["power_kw"].values], dtype=np.float64)
        timestamp = recent_pv_data["timestamp"].values
    else:
        # Make fake PV data; this is where we could add the history of a PV system
        power_kw = np.array([[np.nan]])
        timestamp = [ts]

    da = xr.DataArray(
        data=power_kw,
        dims=["pv_id", "timestamp"],
        coords={
            "longitude": (["pv_id"], [site.longitude]),
            "latitude": (["pv_id"], [site.latitude]),
            "timestamp": timestamp,
            "pv_id": [1],
            "kwp": (["pv_id"], [site.capacity_kwp]),
            "tilt": (["pv_id"], [site.tilt]),
            "orientation": (["pv_id"], [site.orientation]),
        },
    )
    da = da.to_dataset(name="generation_kw")

    return da

def make_pv_data(
        site: PVSite | PVSiteWithInverter,
        ts: pd.Timestamp,
        live_generation: pd.DataFrame | None = None
        ) -> xr.Dataset:
    """
    Make PV data by combining live data from various inverters.

    :param site: the PV site
    :param ts: the timestamp of the site
    :param live_generation: a dataframe containing live generation data for the site
    :return: The combined PV dataset in xarray form
    """
    if isinstance(site, PVSiteWithInverter):
        live_generation_kw = site.get_inverter().get_data(ts)
    else:
        live_generation_kw = live_generation


    # Process the PV data
    da = process_pv_data(live_generation_kw, ts, site)

    return da



import numpy as np
import pandas as pd


def metrics(results_df: pd.DataFrame, pv_metadata: pd.DataFrame, include_night: bool = False):
    """
    Calculate and print metrics: MAE

    There is an option to include nighttime in the calculation of the MAE.

    results_df dataframe with the following columns
    - timestamp
    - pv_id
    - horizon_hours
    - forecast_power
    - generation_power

    pv_metadata is a dataframe with the following columns
    - pv_id
    - capacity

    """

    # remove night time
    if not include_night:
        results_df = results_df[results_df["generation_power"] > 0.1]

    # merge pv_metadata with results_df
    results_df = pd.merge(results_df, pv_metadata, on="pv_id")

    mae = np.round((results_df["forecast_power"] - results_df["generation_power"]).abs().mean(), 4)
    mae_normalized = np.round(
        ((results_df["forecast_power"] - results_df["generation_power"]) / results_df["capacity"])
        .abs()
        .mean(),
        4,
    )
    print(f"MAE: {mae} kw, normalized {100 * mae_normalized} %")

    # calculate metrics over the different horizons hours
    # find all unique horizon_hours
    horizon_hours = results_df["horizon_hour"].unique()
    horizon_groups = [[x, x] for x in horizon_hours]
    horizon_groups += [[3, 4], [5, 8], [9, 16], [17, 24], [24, 48], [0, 36]]

    for horizon_group in horizon_groups:
        horizon_group_df = results_df[
            results_df["horizon_hour"].between(horizon_group[0], horizon_group[1])
        ]
        mae = np.round(
            (horizon_group_df["forecast_power"] - horizon_group_df["generation_power"])
            .abs()
            .mean(),
            3,
        )
        sem = np.round(
            (
                (horizon_group_df["forecast_power"] - horizon_group_df["generation_power"])
                .abs()
                .std()
                / 50**0.5
            ),
            3,
        )

        mae_normalized = np.round(
            (
                (horizon_group_df["forecast_power"] - horizon_group_df["generation_power"])
                / horizon_group_df["capacity"]
            )
            .abs()
            .mean(),
            3,
        )

        print(
            f"MAE for horizon {horizon_group}: {mae} +- {1.96 * sem:.3g}. "
            f"mae_normalized: {100 * mae_normalized:.3g} %"
        )

        # TODO add more metrics using ocf_ml_metrics



from datetime import datetime


def get_file_path(
    latitude: float, longitude: float, capacity_kwp: float, model: str = "gb", time: datetime = None
) -> str:
    return time.strftime(
        f"data/%Y/%-m/%-d/{model}_{latitude}_{longitude}_{capacity_kwp}_%Y%m%d_%H.csv"
    )



from quartz_solar_forecast.utils.forecast_csv import write_out_forecasts
from datetime import datetime

if __name__ == "__main__":
    site_name = input("Enter site name: ")
    start_date_input = input("Enter start date (dd/mm/yyyy): ")
    end_date_input = input("Enter end date (dd/mm/yyyy): ")
    latitude = float(input("Enter latitude: "))
    longitude = float(input("Enter longitude: "))
    capacity_kwp = float(input("Enter capacity in kWp: "))

    start_datetime = datetime.strptime(start_date_input, "%d/%m/%Y").strftime("%Y-%m-%d 00:00:00")
    end_datetime = datetime.strptime(end_date_input, "%d/%m/%Y").strftime("%Y-%m-%d 00:00:00")

    write_out_forecasts(
        init_time_freq=6,
        start_datetime=start_datetime,
        end_datetime=end_datetime,
        site_name=site_name,
        latitude=latitude,
        longitude=longitude,
        capacity_kwp=capacity_kwp
    )


"""Main API."""

import os
from datetime import UTC, datetime
from importlib.metadata import version

import pandas as pd
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from pydantic import BaseModel, Field

from quartz_solar_forecast.forecast import run_forecast
from quartz_solar_forecast.pydantic_models import PVSite

# Get version from environment variable (set in Docker) or fall back to package version
__version__ = os.getenv("QUARTZ_SOLAR_FORECAST_VERSION") or version("quartz_solar_forecast")

description = """
API for [Open Source Quartz Solar Forecast](
    https://github.com/openclimatefix/open-source-quartz-solar-forecast
).

This project aims to provide a free, open-source photovoltaic (PV) forecasting tool
 that is simple to use and works anywhere in the world.

The forecast outputs the expected PV generation (in kW) for up to 48 hours ahead at a single site.

## How it works

- A machine learning model is trained on historical weather and solar generation data from the UK.
- The model has been trained using data from ~25,000 PV sites.
- Forecasts are generated using weather data from open-meteo.com

## Commercial forecasts

Open Climate Fix also provides a commercial PV forecast service.
For more information, please contact: quartz.support@openclimatefix.org

## Example

**Request:**

```bash
curl -X POST "https://open.quartz.solar/forecast/" -H "Content-Type: application/json" -d '{
"site": {
    "latitude": "37.7749",
    "longitude": "-122.4194",
    "capacity_kwp": "5.0",
    "tilt": "30",
    "orientation": "180",
  },
  "timestamp": "2023-08-14T10:00:00Z",
  "live_generation": [
    {
      "timestamp": "2023-08-14T09:45:00Z",
      "generation": 2.5
    },
    {
      "timestamp": "2023-08-14T09:30:00Z",
      "generation": 2.2
    }
  ]
}'
```

Please use UTC timestamps throughout

**Response:**

```json
{
  "timestamp": "2023-08-14 10:00:00",
  "predictions": {
    "power_kw": [values],
  }
}
```
## Want to learn more

We've presented Quartz Solar Forecast at two open source conferences:

- **FOSDEM 2024** (Free and Open source Software Developers' European Meeting):
How we built Open Quartz, our motivation behind it and its impact on aiding organizations
 in resource optimization [Watch the talk](https://www.youtube.com/watch?v=NAZ2VeiN1N8)

- **LF Energy 2024**: Exploring Open Quartz's developments - new models, inverter APIs
 and our Open Source journey at Open Climate Fix
[Watch the talk](https://www.youtube.com/watch?v=YTaq41ztEDg)

And you can always head over to our
[Github page](https://github.com/openclimatefix/open-source-quartz-solar-forecast)

"""


app = FastAPI(description=description, version=__version__, title="Open Quartz Solar Forecast API")

# CORS middleware setup
origins = [
    "*",
]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


class ForecastValues(BaseModel):
    power_kw: dict[datetime, float] = Field(
        ..., description="Dictionary mapping timestamps to power predictions in kW."
    )


class GenerationValue(BaseModel):
    """Generation Value"""

    timestamp: datetime
    generation: float


class ForecastResponse(BaseModel):
    timestamp: datetime = Field(..., description="Timestamp for the forecast.")
    predictions: ForecastValues = Field(..., description="Forecasted power values.")


class ForecastRequest(BaseModel):
    site: PVSite = Field(..., description="PV site information.")
    timestamp: datetime | None = Field(
        None, description="Optional timestamp for the forecast request."
    )
    live_generation: list[GenerationValue] | None = Field(
        None, description="Optional list of live generation values."
    )
    model_config = {
        "json_schema_extra": {
            "example": {
                "site": {
                    "latitude": 51.5072,
                    "longitude": -0.1276,
                    "capacity_kwp": 5.0,
                    "tilt": 30,
                    "orientation": 180,
                },
                "timestamp": "2025-09-12T05:35:48.277Z",
                "live_generation": [
                    {"timestamp": "2025-09-12T05:00:00Z", "generation": 2.5},
                    {"timestamp": "2025-09-12T04:45:00Z", "generation": 2.2},
                ],
            }
        }
    }


@app.post("/forecast/")
def forecast(forecast_request: ForecastRequest) -> ForecastResponse:
    """Get a PV Forecast for a site."""
    site = forecast_request.site
    ts = forecast_request.timestamp if forecast_request.timestamp else datetime.now(UTC).isoformat()

    live_generation = forecast_request.live_generation

    timestamp = pd.Timestamp(ts).tz_localize(None)
    formatted_timestamp = timestamp.strftime("%Y-%m-%d %H:%M:%S")

    # convert live generation to dataframe
    if live_generation is not None and len(live_generation) > 0:
        live_generation_df = pd.DataFrame([g.model_dump() for g in live_generation])
        live_generation_df["timestamp"] = pd.to_datetime(live_generation_df["timestamp"])
        # if timezone, set to UTC and drop it
        if live_generation_df["timestamp"].dt.tz is not None:
            live_generation_df["timestamp"] = live_generation_df["timestamp"].dt.tz_convert("UTC")
            live_generation_df["timestamp"] = live_generation_df["timestamp"].dt.tz_localize(None)

        # make sure timestamps are ordered from earliest to latest
        live_generation_df = live_generation_df.sort_values(by="timestamp")

        live_generation_df.rename(columns={"generation": "power_kw"}, inplace=True)
    else:
        live_generation_df = None

    site_no_live = PVSite(
        latitude=site.latitude,
        longitude=site.longitude,
        capacity_kwp=site.capacity_kwp,
        tilt=site.tilt,
        orientation=site.orientation,
    )

    predictions = run_forecast(site=site_no_live, ts=timestamp, live_generation=live_generation_df)

    response = {
        "timestamp": formatted_timestamp,
        "predictions": predictions.to_dict(),
    }

    return response



import os
import requests
import zipfile
import shutil
from urllib.parse import urlparse

def download_and_process_zip(url, download_dir='../quartz_solar_forecast/dataset/TZ-SAM'):
    # Create download directory if it doesn't exist
    os.makedirs(download_dir, exist_ok=True)
    
    # Extract filename from URL
    parsed_url = urlparse(url)
    filename = os.path.basename(parsed_url.path)
    zip_path = os.path.join(download_dir, filename)
    
    # Download the zip file
    print(f"Downloading {url}...")
    response = requests.get(url, stream=True)
    response.raise_for_status()  # Raise an exception for HTTP errors
    
    with open(zip_path, 'wb') as f:
        for chunk in response.iter_content(chunk_size=8192):
            f.write(chunk)
    print(f"Downloaded to {zip_path}")
    
    # Extract the zip file
    extract_dir = download_dir
    os.makedirs(extract_dir, exist_ok=True)
    
    print(f"Extracting to {extract_dir}...")
    with zipfile.ZipFile(zip_path, 'r') as zip_ref:
        zip_ref.extractall(extract_dir)
    
    # Delete GPKG and PDF files from the extracted contents
    print("Removing GPKG and PDF files...")
    for root, dirs, files in os.walk(extract_dir):
        for file in files:
            if file.lower().endswith(('.gpkg', '.pdf')):
                file_path = os.path.join(root, file)
                os.remove(file_path)
                print(f"Deleted: {file_path}")
    
    # Delete the original zip file
    print(f"Removing original zip file: {zip_path}")
    os.remove(zip_path)
    
    print("Process completed successfully!")

if __name__ == "__main__":
    url = "https://blog.transitionzero.org/hubfs/Data%20Products/TZ-SAM/TZ-SAM%20_%20Solar%20Asset%20Mapper%20-%20Q4%202024.zip"  # Replace with your actual URL
    download_and_process_zip(url)



"""
Forecasts module

This module contains different forecasting models for solar power prediction.
Different models are put in different files like:
- v1.py contains the v1 model
- v2.py contains the v2 model which was developed by Tryolabs
"""

from .v1 import forecast_v1
from .v1_tilt_orientation import forecast_v1_tilt_orientation
from .v2 import TryolabsSolarPowerPredictor

__all__ = ["forecast_v1", "forecast_v1_tilt_orientation", "TryolabsSolarPowerPredictor"]



from datetime import datetime

import pandas as pd
import requests
from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict

from quartz_solar_forecast.inverters.inverter import AbstractInverter


class GivEnergySettings(BaseSettings):
    model_config = SettingsConfigDict(env_file=".env", extra="ignore")

    api_key: str = Field(alias="GIVENERGY_API_KEY")


class GivEnergyInverter(AbstractInverter):
    def __init__(self, settings: GivEnergySettings):
        self.__settings = settings

    def get_data(self, ts: pd.Timestamp) -> pd.DataFrame | None:
        try:
            return get_givenergy_data(self.__settings)
        except Exception as e:
            print(f"Error retrieving GivEnergy data: {e}")
            return None


def get_inverter_serial_number(settings: GivEnergySettings):
    """
    Fetch the inverter serial number from the GivEnergy communication device API.

    :return: Inverter serial number as a string
    """
    api_key = settings.api_key

    if not api_key:
        raise ValueError("GIVENERGY_API_KEY not set in environment variables")

    url = "https://api.givenergy.cloud/v1/communication-device"

    headers = {
        "Authorization": f"Bearer {api_key}",
        "Content-Type": "application/json",
        "Accept": "application/json",
    }

    response = requests.get(url, headers=headers)

    if response.status_code != 200:
        raise Exception(
            f"Communication device API request failed with status code {response.status_code}"
        )

    data = response.json()["data"]
    if not data:
        raise ValueError("No communication devices found")

    inverter_serial_number = data[0]["inverter"]["serial"]
    return inverter_serial_number


def get_givenergy_data(settings: GivEnergySettings):
    """
    Fetch the latest data from the GivEnergy API and return a DataFrame.

    :return: DataFrame with timestamp and power_kw columns
    """
    api_key = settings.api_key

    if not api_key:
        raise ValueError("GIVENERGY_API_KEY not set in environment variables")

    inverter_serial_number = get_inverter_serial_number(settings)

    url = f"https://api.givenergy.cloud/v1/inverter/{inverter_serial_number}/system-data/latest"

    headers = {
        "Authorization": f"Bearer {api_key}",
        "Content-Type": "application/json",
        "Accept": "application/json",
    }

    response = requests.get(url, headers=headers)

    if response.status_code != 200:
        raise Exception(f"System data API request failed with status code {response.status_code}")

    data = response.json()["data"]

    # Process the data
    timestamp = datetime.strptime(data["time"], "%Y-%m-%dT%H:%M:%SZ")
    power_kw = data["solar"]["power"] / 1000  # Convert W to kW

    # Create DataFrame
    df = pd.DataFrame({"timestamp": [timestamp], "power_kw": [power_kw]})

    return df
